@@ -11,10 +11,13 @@ import re
 import hashlib
 import logging
 import shutil
+import random
+import uuid
 import pandas as pd
+from io import BytesIO, StringIO
 from pykemen.utilities import create_api
 from datetime import datetime, timedelta
-from googleapiclient.http import MediaFileUpload
+from googleapiclient.http import MediaFileUpload, HttpError, MediaIoBaseUpload
 
 logger = logging.getLogger("Analytics")
 logger.setLevel(logging.WARNING)
@@ -35,7 +38,7 @@ class Analytics(object):
         REPORT_RE = r"report_[0-9]{4}-[0-9]{2}-[0-9]{2}_[0-9]{4}-[0-9]{2}-[0-9]{2}\.csv"
         UNSAMPLED_REPORT_RE = r'unsampled_report_[0-9]{4}-[0-9]{2}-[0-9]{2}\.csv'
 
-        def __init__(self, path, start_date, end_date, dimensions, metrics, filters, segments, sort, df, unsampled=False, cache=True):
+        def __init__(self,id_,  path, start_date, end_date, dimensions, metrics, filters, segments, sort, df, unsampled=False, cache=True):
             """Init method initialize and create AnalyticsReport class.
 
             Args:
@@ -51,6 +54,7 @@ class Analytics(object):
             Returns:
                 Analytics.AnalyticsReport: with the given configuration.
             """
+            self.id = id_
             self.path = path
             self.report_re = Analytics.AnalyticsReport.UNSAMPLED_REPORT_RE if unsampled else Analytics.AnalyticsReport.REPORT_RE
             self.start_date = start_date
@@ -134,6 +138,7 @@ class Analytics(object):
             "https://www.googleapis.com/auth/analytics.manage.users",
             ]
         self._analyticsService = create_api("analytics", "v3", scope, secrets, credentials)
+        self._uuid = uuid.uuid4()
 
     def get_report(self, unsampled=False, cache=True, **kwargs):
         """Downloads data from Analytics and caches the result. If the data is alredy cached, skips the
@@ -146,8 +151,6 @@ class Analytics(object):
         Returns:
             Analytics.AnalyticsReport
         """
-        last_hit = 0
-        
         
         id_to_hash = ",".join([
             kwargs.get("dimensions", ""), 
@@ -155,6 +158,7 @@ class Analytics(object):
             kwargs.get("filters", ""),
             kwargs.get("segments", ""),
             ])
+        kwargs["quotaUser"] = self._uuid
         id_ = hashlib.md5(id_to_hash.encode('utf8')).hexdigest()
         start_date = kwargs.get('start_date')
         end_date = kwargs.get('end_date')
@@ -164,6 +168,7 @@ class Analytics(object):
         dtypes.update({metric: float for metric in kwargs.get("metrics", "").split(",")})
         rows = []
         data_frames = []
+        report = None
         if cache and not os.path.isdir(Analytics.CACHE_DIR.format(profile=kwargs.get('ids').replace('ga:', ''), id=id_)):
             os.makedirs(Analytics.CACHE_DIR.format(profile=kwargs.get('ids').replace('ga:', ''), id=id_))
         
@@ -175,21 +180,17 @@ class Analytics(object):
                 end_date=end_date
             )
             if not cache or not self._in_cache(kwargs.get('ids').replace('ga:', ''), id_, start_date, end_date):
-                report = self._analyticsService.data().ga().get(**kwargs).execute()
-                last_hit = time.time()
+                request = self._analyticsService.data().ga().get(**kwargs)
+                report = make_request(request)
                 if report.get('containSampledData'):
                     logger.warn("There are sampled results on the report: {dimensions}{metrics} - date: {start_date} to {end_date}".format(
                             dimensions=kwargs.get("dimensions"), metrics=kwargs.get("metrics"), start_date=start_date, end_date=end_date))
                 rows.extend(report.get('rows', []))
                 iteration = 1
                 while report.get('nextLink'):
-                    try:
-                        time.sleep(.1 - (time.time()-last_hit))
-                    except ValueError:
-                        pass
                     kwargs['start_index'] = 1 + kwargs.get('max_results', 1000) * iteration
-                    report = self._analyticsService.data().ga().get(**kwargs).execute()
-                    last_hit = time.time()
+                    request = self._analyticsService.data().ga().get(**kwargs)
+                    report = make_request(request)
                     rows.extend(report.get('rows', []))
                     iteration += 1
                 df = pd.DataFrame(data=rows, columns=columns)
@@ -197,6 +198,8 @@ class Analytics(object):
                 if cache:
                     df.to_csv(filename, index=False, encoding='utf-8')
                     logger.info('Saved file' + filename)
+            else:
+                df = pd.read_csv(filename, index_col=False, dtype=dtypes)
         else:
             startDate = datetime.strptime(start_date, "%Y-%m-%d")
             endDate = datetime.strptime(end_date, "%Y-%m-%d")
@@ -209,20 +212,17 @@ class Analytics(object):
                     kwargs["start_date"] = actualDate
                     kwargs["end_date"] = actualDate
                     kwargs["start_index"] = 1
-                    report = self._analyticsService.data().ga().get(**kwargs).execute()
+                    request = self._analyticsService.data().ga().get(**kwargs)
+                    report = make_request(request)
                     if report.get("containsSampledData"):
                         logger.warn("There are sampled results on the report: {dimensions}{metrics} - date{date}".format(
                             dimensions=kwargs.get("dimensions"), metrics=kwargs.get("metrics"), date=actualDate))
                     rows.extend(report.get("rows", []))
                     iteration = 1
                     while report.get("nextLink"):
-                        try:
-                            time.sleep(.1 - (time.time()-last_hit))
-                        except ValueError:
-                            pass
                         kwargs["start_index"] = 1 + kwargs.get('max_results', 1000) * iteration
-                        report = self._analyticsService.data().ga().get(**kwargs).execute()
-                        last_hit = time.time()
+                        request = self._analyticsService.data().ga().get(**kwargs)
+                        report = make_request(request)
                         rows.extend(report.get("rows", []))
                         iteration += 1
                     df = pd.DataFrame(data=rows, columns=columns)
@@ -231,7 +231,12 @@ class Analytics(object):
                         df.to_csv(filename, index=False, encoding='utf-8')
                         logger.info("Saved file " + filename)
                     rows = []
+                else:
+                    df = pd.read_csv(filename, index_col=False, dtype=dtypes)
+                    data_frames.append(df)
+                
             df = pd.concat(data_frames, ignore_index=True)
+            df = df.astype(dtypes)
             df = df.groupby(kwargs.get("dimensions", "").split(",")).sum().reset_index()
             for sort in kwargs.get("sort", "").split(","):
                 if sort.startswith('-'):
@@ -239,7 +244,7 @@ class Analytics(object):
                 elif sort != '':
                     df = df.sort_values(by=[sort])
         df = df.astype(dtypes)
-        return Analytics.AnalyticsReport(
+        return Analytics.AnalyticsReport(id_,
             Analytics.CACHE_DIR.format(profile=kwargs.get('ids', '').replace('ga:', ''), id=id_),
             start_date,
             end_date,
@@ -253,7 +258,7 @@ class Analytics(object):
             cache=cache
         )
 
-    def data_import(self, accountId, webPropertyId, dataSourceId, filename):
+    def data_import_upload_data(self, accountId, webPropertyId, dataSourceId, filename=None, content=None, dataframe=None):
         """Import a csv to Analytics through a data import.
 
         Args:
@@ -264,24 +269,73 @@ class Analytics(object):
 
         Returns:
             None if the upload succeed, raise an error otherwise"""
-        media = MediaFileUpload(filename, mimetype='application/octet-stream', resumable=False)
-        response = self._analyticsService.management().uploads().uploadData(
+        if filename is None and content is None and dataframe is None:
+            raise Exception("In order to upload data, you have either to introduce a valid filename or a content.")
+        if filename:
+            media = MediaFileUpload(filename, mimetype='application/octet-stream', resumable=False)
+        elif content:
+            media = MediaIoBaseUpload(BytesIO(content.encode("utf-8")), mimetype='application/octet-stream', resumable=False)
+        else:
+            s = StringIO()
+            dataframe.to_csv(s, index=False)
+            media = MediaIoBaseUpload(BytesIO(s.getvalue().encode("utf-8")), mimetype='application/octet-stream', resumable=False)
+        request = self._analyticsService.management().uploads().uploadData(
             accountId=accountId,
             webPropertyId=webPropertyId,
             customDataSourceId=dataSourceId,
-            media_body=media).execute()
+            media_body=media)
+        response = make_request(request)
         while response.get('status') == 'PENDING':
             time.sleep(60)
-            response = self._analyticsService.management().uploads().get(
+            request = self._analyticsService.management().uploads().get(
                 accountId=accountId,
                 webPropertyId=webPropertyId,
                 customDataSourceId=dataSourceId,
                 uploadId=response.get('id')
-            ).execute()
+            )
+            response = make_request(request)
         if response.get('status') == 'FAILED':
             raise Exception(json.dumps(response.get('error'), indent=2))
         if response.get('status') == 'COMPLETED':
             pass
+    
+    def data_import(self, accountId, webPropertyId, dataSourceId, filename=None, content=None):
+        logging.warn("method {c}.data_import is deprecated, pleas use {c}.data_import_upload_data instead.".format(c=self.__class__))
+        self.data_import_upload_data(accountId, webPropertyId, dataSourceId, filename=filename, content=content)
+
+    
+    def data_import_list(self, accountId, webPropertyId, dataSourceId):
+
+        request = self._analyticsService.management().uploads().list(
+            accountId=accountId,
+            webPropertyId=webPropertyId,
+            customDataSourceId=dataSourceId
+        )
+        response = make_request(request)
+        imports = [item.get("id") for item in response.get("items", [])]
+
+        while response.get("nextLink"):
+            request = self._analyticsService.management().uploads().list(
+                accountId=accountId,
+                webPropertyId=webPropertyId,
+                customDataSourceId=dataSourceId,
+                start_index= response.get("start_index") + response.get("itemsPerPage")
+            )
+            response = make_request(request)
+            imports = imports + [item.get("id") for item in response.get("items", [])]
+        return imports
+
+
+    def data_import_delete_data(self, accountId, webPropertyId, dataSourceId, ids):
+        request = self._analyticsService.management().uploads().deleteUploadData(
+            accountId=accountId,
+            webPropertyId=webPropertyId,
+            customDataSourceId=dataSourceId,
+            body={
+                "customDataImportUids": ids
+            }
+        )
+        response = make_request(request)
 
     def _in_cache_by_day(self, profile, id_, date):
         """Check if a specific report is stored in cache.
@@ -346,4 +400,19 @@ def filter_report_files_by_date(filename, start_date, end_date):
     start_date = datetime.strptime(start_date, "%Y-%m-%d")
     end_date = datetime.strptime(end_date, "%Y-%m-%d")
     filename_date = datetime.strptime(filename, "unsampled_report_%Y-%m-%d.csv")
-    return start_date <= filename_date and end_date >= filename_date
\ No newline at end of file
+    return start_date <= filename_date and end_date >= filename_date
+
+
+def make_request(request):
+    for n in range(0, 5):
+        try:
+            return request.execute()
+        except HttpError as error:
+            if error.resp.reason in ['userRateLimitExceeded', 'quotaExceeded',
+                                    'internalServerError', 'backendError']:
+                logger.warn(error.resp.reason)
+                time.sleep((2 ** n) + random.random())
+            else:
+                break
+    
+    return request.execute()
